"""
This module contains functions for concatenating multiple WrappedRedvoxPackets together.
"""

import itertools
import typing

import numpy as _np

import redvox.common.date_time_utils as _date_time_utils
import redvox.api900.exceptions as _exceptions
import redvox.api900.sensors.evenly_sampled_sensor as evenly_sampled_sensor
import redvox.api900.sensors.unevenly_sampled_sensor as unevenly_sampled_sensor

import redvox.api900.sensors.microphone_sensor as _microphone_sensor
import redvox.api900.sensors.barometer_sensor as _barometer_sensor
import redvox.api900.sensors.location_sensor as _location_sensor
import redvox.api900.sensors.gyroscope_sensor as _gyroscope_sensor
import redvox.api900.sensors.magnetometer_sensor as _magnetometer_sensor
import redvox.api900.sensors.accelerometer_sensor as _accelerometer_sensor
import redvox.api900.sensors.light_sensor as _light_sensor
import redvox.api900.sensors.infrared_sensor as _infrared_sensor
import redvox.api900.sensors.image_sensor as _image_sensor
import redvox.api900.sensors.time_synchronization_sensor as _time_synchronization_sensor
from redvox.api900.wrapped_redvox_packet import WrappedRedvoxPacket

# pylint: disable=C0103
RedvoxSensor = typing.Union[
    evenly_sampled_sensor.EvenlySampledSensor,
    unevenly_sampled_sensor.UnevenlySampledSensor,
    _microphone_sensor.MicrophoneSensor,
    _barometer_sensor.BarometerSensor,
    _location_sensor.LocationSensor,
    _time_synchronization_sensor.TimeSynchronizationSensor,
    _accelerometer_sensor.AccelerometerSensor,
    _gyroscope_sensor.GyroscopeSensor,
    _magnetometer_sensor.MagnetometerSensor,
    _light_sensor.LightSensor,
    _infrared_sensor.InfraredSensor,
    _image_sensor.ImageSensor
]
# pylint: disable=C0103
RedvoxSensors = typing.List[RedvoxSensor]


_NONE_HASH = hash(None)


def _partial_hash_sensor(sensor: typing.Optional[RedvoxSensor]) -> int:
    """
    Performs a partial hash on a sensor hashing the sensor name, sample rate, and payload type.
    :param sensor: The sensor to hash.
    :return: Hash of the sensor.
    """
    if sensor is None:
        return _NONE_HASH

    if isinstance(sensor, unevenly_sampled_sensor.UnevenlySampledSensor):
        return hash((sensor.sensor_name(), sensor.payload_type()))

    if isinstance(sensor, evenly_sampled_sensor.EvenlySampledSensor):
        return hash((sensor.sample_rate_hz(), sensor.sensor_name(), sensor.payload_type()))

    if isinstance(sensor, _time_synchronization_sensor.TimeSynchronizationSensor):
        return hash("TimeSynchronizationSensor")

    raise _exceptions.ConcatenationException("trying to hash non-sensor type=%s" % type(sensor))


def _partial_hash_packet(wrapped_redvox_packet) -> int:
    """
    Computes the partial hash of a wrapped redvox packet.

    The hash is computed by hashing all of the partial hashes of the sensor channels.
    :param wrapped_redvox_packet: Packet to hash.
    :return: The has of this packet.
    """
    if wrapped_redvox_packet is None:
        return _NONE_HASH

    return hash((wrapped_redvox_packet.redvox_id(),
                 wrapped_redvox_packet.uuid(),
                 _partial_hash_sensor(wrapped_redvox_packet.microphone_sensor()),
                 _partial_hash_sensor(wrapped_redvox_packet.barometer_sensor()),
                 _partial_hash_sensor(wrapped_redvox_packet.location_sensor()),
                 _partial_hash_sensor(wrapped_redvox_packet.time_synchronization_sensor()),
                 _partial_hash_sensor(wrapped_redvox_packet.accelerometer_sensor()),
                 _partial_hash_sensor(wrapped_redvox_packet.gyroscope_sensor()),
                 _partial_hash_sensor(wrapped_redvox_packet.magnetometer_sensor()),
                 _partial_hash_sensor(wrapped_redvox_packet.light_sensor()),
                 _partial_hash_sensor(wrapped_redvox_packet.infrared_sensor())))


def _packet_len_s(wrapped_redvox_packet) -> float:
    """
    Returns the length of a packet in seconds.
    :param wrapped_redvox_packet: Packet to find the length of.
    :return: The length of a packet in seconds.
    """
    microphone_sensor = wrapped_redvox_packet.microphone_sensor()
    return len(microphone_sensor.payload_values()) / microphone_sensor.sample_rate_hz()


def _identify_gaps(wrapped_redvox_packets,
                   allowed_timing_error_s: float) -> typing.List[int]:
    """
    Identifies discontinuities in sensor data by checking if sensors drop in and out and by comparing timing info.
    :param wrapped_redvox_packets: Packets to look for gaps in.
    :param allowed_timing_error_s: The amount of timing error in seconds.
    :return: A list of indices into the original list where gaps were found.
    """

    if len(wrapped_redvox_packets) <= 1:
        return []

    gaps = set()

    truth_len = _packet_len_s(wrapped_redvox_packets[0])
    for i in range(1, len(wrapped_redvox_packets)):
        prev_packet = wrapped_redvox_packets[i - 1]
        next_packet = wrapped_redvox_packets[i]

        # Sensor discontinuity
        prev_sensors = [prev_packet.has_microphone_sensor(),
                        prev_packet.has_barometer_sensor(),
                        prev_packet.has_time_synchronization_sensor(),
                        prev_packet.has_accelerometer_sensor(),
                        prev_packet.has_gyroscope_sensor(),
                        prev_packet.has_infrared_sensor(),
                        prev_packet.has_light_sensor(),
                        prev_packet.has_image_sensor(),
                        prev_packet.has_location_sensor(),
                        prev_packet.has_magnetometer_sensor()]

        next_sensors = [next_packet.has_microphone_sensor(),
                        next_packet.has_barometer_sensor(),
                        next_packet.has_time_synchronization_sensor(),
                        next_packet.has_accelerometer_sensor(),
                        next_packet.has_gyroscope_sensor(),
                        next_packet.has_infrared_sensor(),
                        next_packet.has_light_sensor(),
                        next_packet.has_image_sensor(),
                        next_packet.has_location_sensor(),
                        next_packet.has_magnetometer_sensor()]

        # pylint: disable=C0200
        for j in range(len(prev_sensors)):
            if prev_sensors[j] != next_sensors[j]:
                gaps.add(i)

        # Time based gaps
        prev_timestamp = prev_packet.microphone_sensor().first_sample_timestamp_epoch_microseconds_utc()
        next_timestamp = next_packet.microphone_sensor().first_sample_timestamp_epoch_microseconds_utc()
        print(next_timestamp - prev_timestamp)
        if _date_time_utils.microseconds_to_seconds(next_timestamp - prev_timestamp) > (
                truth_len + allowed_timing_error_s):
            gaps.add(i)
            print("time gap")
            truth_len = _packet_len_s(wrapped_redvox_packets[i])
        prev_mach_time_zero = prev_packet.mach_time_zero()
        next_mach_time_zero = next_packet.mach_time_zero()
        if next_mach_time_zero != prev_mach_time_zero:
            gaps.add(i)
            print("mach time zero gap")

    return sorted(list(gaps))


def _identify_sensor_changes(wrapped_redvox_packets: typing.List) -> typing.List[int]:
    """
    Identifies discontinuities in sensors including sample ratechange , redvox id/uuid change, and sensor name change.
    :param wrapped_redvox_packets: A list of WrappedRedvoxPackets
    :return: A list of indexes where sensor changes were detected.
    """
    if len(wrapped_redvox_packets) <= 1:
        return []

    sensor_changes: typing.List[int] = []
    truth_hash = _partial_hash_packet(wrapped_redvox_packets[0])
    for i in range(1, len(wrapped_redvox_packets)):
        next_packet = wrapped_redvox_packets[i]

        candidate_hash = _partial_hash_packet(next_packet)
        if truth_hash != candidate_hash:
            sensor_changes.append(i)
            truth_hash = candidate_hash

    return sensor_changes


def _concat_numpy(sensors: RedvoxSensors,
                  array_extraction_fn: typing.Callable[[RedvoxSensor], _np.ndarray]) -> _np.ndarray:
    """
    Given a list of sensors concatenate the numpy arrays found with the extraction function.
    :param sensors: Sensors to extract arrays from.
    :param array_extraction_fn: A function that takes a sensor and returns a numpy array.
    :return: An array of concatenated arrays.
    """
    return _np.concatenate(list(map(array_extraction_fn, sensors)))


def _concat_lists(sensors: RedvoxSensors,
                  list_extraction_fn: typing.Callable[[RedvoxSensor], typing.List[str]]) -> typing.List[
                      str]:
    """
    Given a list of sensors concatenate the lists found with the extraction function.
    :param sensors: Sensors to extract lists from.
    :param list_extraction_fn: A function that takes a sensor and returns a list.
    :return: A list of concatenated arrays.
    """
    metadata_list = list(map(list_extraction_fn, sensors))
    return list(itertools.chain(*metadata_list))


def _concat_continuous_data(wrapped_redvox_packets: typing.List[WrappedRedvoxPacket]) -> WrappedRedvoxPacket:
    """
    Given a set of continuous wrapped redvox packets, concatenate the packets together by concatting the timestamps,
    payload values, and metadata.
    :param wrapped_redvox_packets: Packets to concatenate.
    :return: A single WrappedRedvoxPacket with concatenated data.
    """
    first_packet = wrapped_redvox_packets[0]

    # Concat channels
    if first_packet.has_microphone_sensor():
        sensors = list(map(WrappedRedvoxPacket.microphone_sensor, wrapped_redvox_packets))
        sensors[0].set_payload_values(_concat_numpy(sensors, _microphone_sensor.MicrophoneSensor.payload_values)) \
            .set_metadata(_concat_lists(sensors, _microphone_sensor.MicrophoneSensor.metadata))

    if first_packet.has_barometer_sensor():
        sensors = list(map(WrappedRedvoxPacket.barometer_sensor, wrapped_redvox_packets))
        sensors[0].set_payload_values(_concat_numpy(sensors, _barometer_sensor.BarometerSensor.payload_values)) \
            .set_timestamps_microseconds_utc(
                _concat_numpy(sensors, _barometer_sensor.BarometerSensor.timestamps_microseconds_utc)) \
            .set_metadata(_concat_lists(sensors, _barometer_sensor.BarometerSensor.metadata))

    if first_packet.has_location_sensor():
        sensors = list(map(WrappedRedvoxPacket.location_sensor, wrapped_redvox_packets))
        sensors[0].set_payload_values(
            _concat_numpy(sensors, _location_sensor.LocationSensor.payload_values_latitude),
            _concat_numpy(sensors, _location_sensor.LocationSensor.payload_values_longitude),
            _concat_numpy(sensors, _location_sensor.LocationSensor.payload_values_altitude),
            _concat_numpy(sensors, _location_sensor.LocationSensor.payload_values_speed),
            _concat_numpy(sensors, _location_sensor.LocationSensor.payload_values_accuracy)
        ) \
            .set_timestamps_microseconds_utc(
                _concat_numpy(sensors, _location_sensor.LocationSensor.timestamps_microseconds_utc)) \
            .set_metadata(_concat_lists(sensors, _location_sensor.LocationSensor.metadata))

    if first_packet.has_time_synchronization_sensor():
        sensors = list(map(WrappedRedvoxPacket.time_synchronization_sensor, wrapped_redvox_packets))
        sensors[0].set_payload_values(_concat_numpy(
            sensors,
            _time_synchronization_sensor.TimeSynchronizationSensor.payload_values)) \
            .set_metadata(_concat_lists(sensors, _time_synchronization_sensor.TimeSynchronizationSensor.metadata))

    if first_packet.has_magnetometer_sensor():
        sensors = list(map(WrappedRedvoxPacket.magnetometer_sensor, wrapped_redvox_packets))
        sensors[0].set_payload_values(
            _concat_numpy(sensors, _magnetometer_sensor.MagnetometerSensor.payload_values_x),
            _concat_numpy(sensors, _magnetometer_sensor.MagnetometerSensor.payload_values_y),
            _concat_numpy(sensors, _magnetometer_sensor.MagnetometerSensor.payload_values_z)
        ) \
            .set_timestamps_microseconds_utc(
                _concat_numpy(sensors, _magnetometer_sensor.MagnetometerSensor.timestamps_microseconds_utc)) \
            .set_metadata(_concat_lists(sensors, _magnetometer_sensor.MagnetometerSensor.metadata))

    if first_packet.has_accelerometer_sensor():
        sensors = list(map(WrappedRedvoxPacket.accelerometer_sensor, wrapped_redvox_packets))
        sensors[0].set_payload_values(
            _concat_numpy(sensors, _accelerometer_sensor.AccelerometerSensor.payload_values_x),
            _concat_numpy(sensors, _accelerometer_sensor.AccelerometerSensor.payload_values_y),
            _concat_numpy(sensors, _accelerometer_sensor.AccelerometerSensor.payload_values_z)
        ) \
            .set_timestamps_microseconds_utc(
                _concat_numpy(sensors, _accelerometer_sensor.AccelerometerSensor.timestamps_microseconds_utc)) \
            .set_metadata(_concat_lists(sensors, _accelerometer_sensor.AccelerometerSensor.metadata))

    if first_packet.has_gyroscope_sensor():
        sensors = list(map(WrappedRedvoxPacket.gyroscope_sensor, wrapped_redvox_packets))
        sensors[0].set_payload_values(
            _concat_numpy(sensors, _gyroscope_sensor.GyroscopeSensor.payload_values_x),
            _concat_numpy(sensors, _gyroscope_sensor.GyroscopeSensor.payload_values_y),
            _concat_numpy(sensors, _gyroscope_sensor.GyroscopeSensor.payload_values_z)
        ) \
            .set_timestamps_microseconds_utc(
                _concat_numpy(sensors, _gyroscope_sensor.GyroscopeSensor.timestamps_microseconds_utc)) \
            .set_metadata(_concat_lists(sensors, _gyroscope_sensor.GyroscopeSensor.metadata))

    if first_packet.has_light_sensor():
        sensors = list(map(WrappedRedvoxPacket.light_sensor, wrapped_redvox_packets))
        sensors[0].set_payload_values(_concat_numpy(sensors, _light_sensor.LightSensor.payload_values)) \
            .set_timestamps_microseconds_utc(
                _concat_numpy(sensors, _light_sensor.LightSensor.timestamps_microseconds_utc)) \
            .set_metadata(_concat_lists(sensors, _light_sensor.LightSensor.metadata))

    if first_packet.has_infrared_sensor():
        sensors = list(map(WrappedRedvoxPacket.infrared_sensor, wrapped_redvox_packets))
        sensors[0].set_payload_values(_concat_numpy(sensors, _infrared_sensor.InfraredSensor.payload_values)) \
            .set_timestamps_microseconds_utc(
                _concat_numpy(sensors, _infrared_sensor.InfraredSensor.timestamps_microseconds_utc)) \
            .set_metadata(_concat_lists(sensors, _infrared_sensor.InfraredSensor.metadata))

    # Concat metadata
    all_metadata = list(map(WrappedRedvoxPacket.metadata, wrapped_redvox_packets))
    first_packet.set_metadata(list(itertools.chain(*all_metadata)))

    return first_packet


def concat_wrapped_redvox_packets(wrapped_redvox_packets: typing.List[WrappedRedvoxPacket]) \
        -> typing.List[WrappedRedvoxPacket]:
    """
    Concatenates multiple WrappedRedvoxPackets together by combining sensor timestamps, values, and metadata.
    :param wrapped_redvox_packets: Packets to concatenate.
    :return: A list of concatenated WrappedRedvoxPackets. A list is returned because each packet represents a single
             continuous range of data.
    """
    if wrapped_redvox_packets is None or len(wrapped_redvox_packets) == 0:
        return []

    if len(wrapped_redvox_packets) == 1:
        return wrapped_redvox_packets

    # Check that packets are from same device
    device_ids = set(map(lambda packet: packet.redvox_id(), wrapped_redvox_packets))
    if len(device_ids) != 1:
        raise _exceptions.ConcatenationException("Not all packets from same device %s" % str(device_ids))

    # Check that packets are ordered
    machine_times = list(map(lambda packet: packet.app_file_start_timestamp_machine(),
                             wrapped_redvox_packets))

    if not _np.all(_np.diff(_np.array(machine_times)) > 0):
        raise _exceptions.ConcatenationException("Packets are not strictly monotonic")

    # Identify gaps
    gaps = _identify_gaps(wrapped_redvox_packets, 5)

    # Concat
    concatenated_packets = []
    start = 0
    end = len(wrapped_redvox_packets)

    for gap_idx in gaps:
        concatenated_packets.append(_concat_continuous_data(wrapped_redvox_packets[start:gap_idx]))
        start = gap_idx

    concatenated_packets.append(_concat_continuous_data(wrapped_redvox_packets[start:end]))

    return concatenated_packets



"""
This module contains API definitions and functions for working with raw RedVox data packets.
"""

from dataclasses import dataclass
from typing import Callable, List, Optional, Tuple

from dataclasses_json import dataclass_json
import requests

from redvox.cloud.api import ApiConfig, post_req
import redvox.cloud.data_io as data_io
from redvox.cloud.routes import RoutesV1


@dataclass_json
@dataclass
class ReportDataReq:
    """
    A request for a signed URL to a RedVox report distribution.
    """
    auth_token: str
    report_id: str
    secret_token: Optional[str] = None


@dataclass_json
@dataclass
class ReportDataResp:
    """
    Response for a report signed URL.
    """
    signed_url: str

    def download_buf(self, retries: int = 3) -> Optional[bytes]:
        """
        Downloads the zip compressed report data and returns it as a byte buffer.
        :param retries: Number of times to retry the download of failure.
        :return: Bytes of the compressed report data.
        """
        return data_io.get_file(self.signed_url, retries)

    def download_fs(self, out_dir: str, retries: int = 3) -> Tuple[str, int]:
        """
        Downloads the zip compressed report data to the file system.
        :param out_dir: Output directory to download data to.
        :param retries: Number of times to retry the download on failure.
        :return: A tuple containing the file name and the number of bytes written.
        """
        return data_io.download_file(self.signed_url, requests.Session(), out_dir, retries)


@dataclass_json
@dataclass
class DataRangeReq:
    """
    Definition of data range request.
    """
    auth_token: str
    start_ts_s: int
    end_ts_s: int
    redvox_ids: List[str]
    secret_token: Optional[str] = None


@dataclass_json
@dataclass
class DataRangeResp:
    """
    Definition of a data range response.
    """
    signed_urls: List[str]

    def download_fs(self, out_dir: str, retries: int = 3) -> None:
        """
        Download the referenced packets to the provided output directory.
        :param out_dir: Output directory to store the downloaded files.
        :param retries: Number of times to retry downloading the file on failure.
        """
        data_io.download_files_parallel(self.signed_urls, out_dir, retries)


def request_report_data(api_config: ApiConfig,
                        report_data_req: ReportDataReq,
                        session: Optional[requests.Session] = None,
                        timeout: Optional[float] = None) -> Optional[ReportDataResp]:
    """
    Makes an API call to generate a signed URL of a RedVox report.
    :param api_config: An API config.
    :param report_data_req: The request.
    :param session: An (optional) session for re-using an HTTP client.
    :return: The response.
    """
    # noinspection Mypy
    handle_resp: Callable[[requests.Response], ReportDataResp] = lambda resp: ReportDataResp.from_dict(resp.json())
    return post_req(api_config,
                    RoutesV1.DATA_REPORT_REQ,
                    report_data_req,
                    handle_resp,
                    session,
                    timeout)


def request_range_data(api_config: ApiConfig,
                       data_range_req: DataRangeReq,
                       session: Optional[requests.Session] = None,
                       timeout: Optional[float] = None) -> DataRangeResp:
    """
    Requests signed URLs for a range of RedVox packets.
    :param api_config: An API config.
    :param data_range_req: The request.
    :param session: An (optional) session for re-using an HTTP client.
    :return: A DataRangeResp.
    """
    # noinspection Mypy
    handle_resp: Callable[[requests.Response], DataRangeResp] = lambda resp: DataRangeResp.from_dict(resp.json())
    res: Optional[DataRangeResp] = post_req(api_config,
                                            RoutesV1.DATA_RANGE_REQ,
                                            data_range_req,
                                            handle_resp,
                                            session,
                                            timeout)

    return res if res else DataRangeResp([])



"""
This module contains classes and methods for working with location sensors.
"""

import typing

import numpy

import redvox.api900.constants as constants
import redvox.api900.lib.api900_pb2 as api900_pb2
from redvox.api900.sensors.unevenly_sampled_channel import UnevenlySampledChannel
from redvox.api900.sensors.unevenly_sampled_sensor import UnevenlySampledSensor


# pylint: disable=R0904
class LocationSensor(UnevenlySampledSensor):
    """
    High-level wrapper around location channels.
    """

    def __init__(self, unevenly_sampled_channel: typing.Optional[UnevenlySampledChannel] = None):
        """
        Initialize this channel
        :param unevenly_sampled_channel: Instance of UnevenlySampledChannel containing location data
        """
        super().__init__(unevenly_sampled_channel)
        self._unevenly_sampled_channel.set_channel_types([
            api900_pb2.LATITUDE,
            api900_pb2.LONGITUDE,
            api900_pb2.ALTITUDE,
            api900_pb2.SPEED,
            api900_pb2.ACCURACY
        ])

    def _payload_values(self):
        """
        Return the location payload as an interleaved payload with the following format:
        [[latitude_0, longitude_0, altitude_0, speed_0, accuracy_0],
         [latitude_1, longitude_1, altitude_1, speed_1, accuracy_1],
         ...,
         [latitude_n, longitude_n, altitude_n, speed_n, accuracy_n],]
        :return: array containing interleaved values of the 5 channels
        """
        return self._unevenly_sampled_channel.get_multi_payload([
            api900_pb2.LATITUDE,
            api900_pb2.LONGITUDE,
            api900_pb2.ALTITUDE,
            api900_pb2.SPEED,
            api900_pb2.ACCURACY
        ])

    # pylint: disable=R0913
    def set_payload_values(self,
                           latitude_payload: typing.Union[typing.List[float], numpy.ndarray],
                           longitude_payload: typing.Union[typing.List[float], numpy.ndarray],
                           altitude_payload: typing.Union[typing.List[float], numpy.ndarray],
                           speed_payload: typing.Union[typing.List[float], numpy.ndarray],
                           accuracy_payload: typing.Union[typing.List[float], numpy.ndarray]) -> 'LocationSensor':
        """
        Sets the location sensor's payload. Note that if one of the below channels don't exist, you must still provide
        a zero-filled payload that is the same length as all other paylods.
        :param latitude_payload: The latitude payload.
        :param longitude_payload: The longitude payload.
        :param altitude_payload: The altitude payload.
        :param speed_payload: The speed payload
        :param accuracy_payload: The accuracy payload.
        :return: An instance of the sensor.
        """
        self._unevenly_sampled_channel.set_interleaved_payload([latitude_payload,
                                                                longitude_payload,
                                                                altitude_payload,
                                                                speed_payload,
                                                                accuracy_payload],
                                                               constants.PayloadType.FLOAT64_PAYLOAD)
        return self

    def payload_values_latitude(self):
        """
        Returns the latitude component of this channel's payload.
        :return: The latitude component of this channel's payload.
        """
        return self._unevenly_sampled_channel.get_payload(api900_pb2.LATITUDE)

    def payload_values_longitude(self):
        """
        Returns the longitude component of this channel's payload.
        :return: The longitude component of this channel's payload.
        """
        return self._unevenly_sampled_channel.get_payload(api900_pb2.LONGITUDE)

    def payload_values_altitude(self):
        """
        Returns the altitude component of this channel's payload.
        :return: The altitude component of this channel's payload.
        """
        return self._unevenly_sampled_channel.get_payload(api900_pb2.ALTITUDE)

    def payload_values_speed(self):
        """
        Returns the speed component of this channel's payload.
        :return: The speed component of this channel's payload.
        """
        return self._unevenly_sampled_channel.get_payload(api900_pb2.SPEED)

    def payload_values_accuracy(self):
        """
        Returns the accuracy component of this channel's payload.
        :return: The accuracy component of this channel's payload.
        """
        return self._unevenly_sampled_channel.get_payload(api900_pb2.ACCURACY)

    def payload_values_latitude_mean(self) -> float:
        """
        Returns the mean latitude component of this channel's payload.
        :return: The mean latitude component of this channel's payload.
        """
        return self._unevenly_sampled_channel.get_value_mean(api900_pb2.LATITUDE)

    def payload_values_longitude_mean(self) -> float:
        """
        Returns the mean longitude component of this channel's payload.
        :return: The mean longitude component of this channel's payload.
        """
        return self._unevenly_sampled_channel.get_value_mean(api900_pb2.LONGITUDE)

    def payload_values_altitude_mean(self) -> float:
        """
        Returns the mean altitude component of this channel's payload.
        :return: The mean altitude component of this channel's payload.
        """
        return self._unevenly_sampled_channel.get_value_mean(api900_pb2.ALTITUDE)

    def payload_values_speed_mean(self) -> float:
        """
        Returns the mean speed component of this channel's payload.
        :return: The mean speed component of this channel's payload.
        """
        return self._unevenly_sampled_channel.get_value_mean(api900_pb2.SPEED)

    def payload_values_accuracy_mean(self) -> float:
        """
        Returns the mean accuracy component of this channel's payload.
        :return: The mean accuracy component of this channel's payload.
        """
        return self._unevenly_sampled_channel.get_value_mean(api900_pb2.ACCURACY)

    def payload_values_latitude_median(self) -> float:
        """
        Returns the median latitude component of this channel's payload.
        :return: The median latitude component of this channel's payload.
        """
        return self._unevenly_sampled_channel.get_value_median(api900_pb2.LATITUDE)

    def payload_values_longitude_median(self) -> float:
        """
        Returns the median longitude component of this channel's payload.
        :return: The median longitude component of this channel's payload.
        """
        return self._unevenly_sampled_channel.get_value_median(api900_pb2.LONGITUDE)

    def payload_values_altitude_median(self) -> float:
        """
        Returns the median altitude component of this channel's payload.
        :return: The median altitude component of this channel's payload.
        """
        return self._unevenly_sampled_channel.get_value_median(api900_pb2.ALTITUDE)

    def payload_values_speed_median(self) -> float:
        """
        Returns the median speed component of this channel's payload.
        :return: The median speed component of this channel's payload.
        """
        return self._unevenly_sampled_channel.get_value_median(api900_pb2.SPEED)

    def payload_values_accuracy_median(self) -> float:
        """
        Returns the median accuracy component of this channel's payload.
        :return: The median accuracy component of this channel's payload.
        """
        return self._unevenly_sampled_channel.get_value_median(api900_pb2.ACCURACY)

    def payload_values_latitude_std(self) -> float:
        """
        Returns the standard deviation latitude component of this channel's payload.
        :return: The standard deviation latitude component of this channel's payload.
        """
        return self._unevenly_sampled_channel.get_value_std(api900_pb2.LATITUDE)

    def payload_values_longitude_std(self) -> float:
        """
        Returns the standard deviation longitude component of this channel's payload.
        :return: The standard deviation longitude component of this channel's payload.
        """
        return self._unevenly_sampled_channel.get_value_std(api900_pb2.LONGITUDE)

    def payload_values_altitude_std(self) -> float:
        """
        Returns the standard deviation altitude component of this channel's payload.
        :return: The standard deviation altitude component of this channel's payload.
        """
        return self._unevenly_sampled_channel.get_value_std(api900_pb2.ALTITUDE)

    def payload_values_speed_std(self) -> float:
        """
        Returns the standard deviation speed component of this channel's payload.
        :return: The standard deviation speed component of this channel's payload.
        """
        return self._unevenly_sampled_channel.get_value_std(api900_pb2.SPEED)

    def payload_values_accuracy_std(self) -> float:
        """
        Returns the standard deviation accuracy component of this channel's payload.
        :return: The standard deviation accuracy component of this channel's payload.
        """
        return self._unevenly_sampled_channel.get_value_std(api900_pb2.ACCURACY)



"""
This module contains custom error definitions used within this SDK.
"""


class RedVoxError(Exception):
    """
    This class represents generic RedVox SDK errors.
    """
    def __init__(self, message: str):
        super().__init__(f"RedVoxError: {message}")



"""
This module contains functions and classes for summarizing ranges of redvox data.
"""

import datetime
import itertools
import typing

import numpy
import matplotlib.pyplot as plt

import redvox.api900.types as types
import redvox.common.date_time_utils as date_time_utils


def _sensor_name(sensor: typing.Optional[types.RedvoxSensor]) -> str:
    """
    Extracts the sensor name from a redvox sensor.
    :param sensor: Sensor to get sensor name for.
    :return: The sensor name for the given sensor.
    """
    if sensor is None:
        return ""

    if isinstance(sensor, types.time_synchronization_sensor.TimeSynchronizationSensor):
        return "time synch"

    return sensor.sensor_name()


def _sensor_samples(sensor: typing.Optional[types.RedvoxSensor]) -> int:
    """
    Returns the number of samples a sensor contains.
    :param sensor: Sensor to test.
    :return: Number of samples in sensor.
    """
    if sensor is None:
        return 0

    if isinstance(sensor, (types.microphone_sensor.MicrophoneSensor,
                           types.time_synchronization_sensor.TimeSynchronizationSensor)):
        return len(sensor.payload_values())

    return len(sensor.timestamps_microseconds_utc())


# pylint: disable=R0903
class SensorSummary:
    """
    This class provides a summary of a sensor.
    """

    def __init__(self, sensor: typing.Optional[types.RedvoxSensor]):
        self.has_sensor = sensor is not None
        self.sensor_name = _sensor_name(sensor)
        self.num_samples = _sensor_samples(sensor)

    def __str__(self):
        if not self.has_sensor:
            return "None"
        return "{SensorSummary: %s with %d samples} " % (self.sensor_name, self.num_samples)


# pylint: disable=R0902
# pylint: disable=R0903
class WrappedRedvoxPacketSummary:
    """
    This class provides a summary of a WrappedRedvoxPacket.
    """

    def __init__(self, redvox_packet):
        self.redvox_id = redvox_packet.redvox_id()
        self.uuid = redvox_packet.uuid()
        self.redvox_id_uuid: str = "%s:%s" % (self.redvox_id, self.uuid)
        self.start_timestamp_us: int = redvox_packet.start_timestamp_us_utc()
        self.end_timestamp_us: int = redvox_packet.end_timestamp_us_utc()
        self.duration_s: float = redvox_packet.duration_s()
        self.sample_rate_hz = redvox_packet.microphone_sensor().sample_rate_hz()
        self.microphone_sensor_summary: SensorSummary = SensorSummary(redvox_packet.microphone_sensor())
        self.barometer_sensor_summary: SensorSummary = SensorSummary(redvox_packet.barometer_sensor())
        self.location_sensor_summary: SensorSummary = SensorSummary(redvox_packet.location_sensor())
        # pylint: disable=C0103
        self.time_synchronization_sensor_summary: SensorSummary = SensorSummary(
            redvox_packet.time_synchronization_sensor())
        self.accelerometer_sensor_summary: SensorSummary = SensorSummary(redvox_packet.accelerometer_sensor())
        self.magnetometer_sensor_summary: SensorSummary = SensorSummary(redvox_packet.magnetometer_sensor())
        self.gyroscope_sensor_summary: SensorSummary = SensorSummary(redvox_packet.gyroscope_sensor())
        self.light_sensor_summary: SensorSummary = SensorSummary(redvox_packet.light_sensor())
        self.infrared_sensor_summary: SensorSummary = SensorSummary(redvox_packet.infrared_sensor())

    def __str__(self):
        return "{WrappedRedvoxPacketSummary: %s %d-%d [%f sec] %f %s %s %s %s %s %s %s %s %s} " % (
            self.redvox_id_uuid,
            self.start_timestamp_us,
            self.end_timestamp_us,
            self.duration_s,
            self.sample_rate_hz,
            str(self.microphone_sensor_summary),
            str(self.barometer_sensor_summary),
            str(self.location_sensor_summary),
            str(self.time_synchronization_sensor_summary),
            str(self.accelerometer_sensor_summary),
            str(self.magnetometer_sensor_summary),
            str(self.gyroscope_sensor_summary),
            str(self.light_sensor_summary),
            str(self.infrared_sensor_summary)
        )


# Type vars for compose function.
# pylint: disable=C0103
X = typing.TypeVar("X")
# pylint: disable=C0103
Y = typing.TypeVar("Y")
# pylint: disable=C0103
Z = typing.TypeVar("Z")


def _compose(fn1: typing.Callable[[X], Y],
             fn2: typing.Callable[[Y], Z]) -> typing.Callable[[X], Z]:
    """
    Composes two functions together.
    :param fn1: A function from X -> Y
    :param fn2: A function from Y -> Z
    :return: A function from X -> Z
    """
    return lambda value: fn2(fn1(value))


def summarize_data(grouped_redvox_packets: typing.Dict[str, typing.List]) -> typing.Dict[
        str, typing.List[WrappedRedvoxPacketSummary]]:
    """
    Given grouped redvox packets, generate summaries for each device and each packet.
    :param grouped_redvox_packets: Grouped packets to generate summary for.
    :return: A dictionary which maps redvox ids to summaries of redvox packets for that device.
    """
    summarized = {}
    for id_uuid, wrapped_redvox_packets in grouped_redvox_packets.items():
        summarized[id_uuid] = list(map(WrappedRedvoxPacketSummary, wrapped_redvox_packets))
    return summarized


def _subsample(values: typing.Union[numpy.ndarray, typing.List], num_samples: int) -> typing.Union[
        numpy.ndarray, typing.List]:
    """
    Returns evenly sampled values from an array.
    From https://stackoverflow.com/a/50685454 and https://stackoverflow.com/a/9873935
    :param values: Sequence to extract evenly sampled samples from.
    :param num_samples: Number of samples to extract.
    :return: A new, subsampled array.
    """
    indexes = numpy.round(numpy.linspace(0, len(values) - 1, num_samples)).astype(int)
    if isinstance(values, numpy.ndarray):
        return numpy.take(values, indexes)
    elif isinstance(values, list):
        return [values[i] for i in indexes]
    else:
        raise RuntimeError("Non sequence type passed to subsample")


# pylint: disable=R0914
def plot_summarized_data(summarized_data: typing.Dict[str, typing.List[WrappedRedvoxPacketSummary]]):
    """
    Given summarized data, plot the data to display sensor activity for all passed in summarized data.
    :param summarized_data: Data to plot.
    """
    all_summaries = list(itertools.chain(*summarized_data.values()))

    start_timestamp_us = numpy.min(
        list(map(lambda packet_summary: packet_summary.start_timestamp_us, all_summaries)))
    end_timestamp_us = numpy.max(list(map(lambda packet_summary: packet_summary.end_timestamp_us, all_summaries)))
    start_s = int(date_time_utils.microseconds_to_seconds(start_timestamp_us))
    end_s = int(date_time_utils.microseconds_to_seconds(end_timestamp_us))
    timestamps = numpy.arange(start_s, end_s + 1)

    # Setup the plot
    fig, axes = plt.subplots(nrows=1)
    axes.set_xlim([timestamps[0], timestamps[-1]])
    axes.set_ylim([0, len(summarized_data.keys()) + 1])

    # Not for each device, let's plot its ranges
    yticks = []
    ylabels = []
    i = 1
    for redvox_id, summaries in summarized_data.items():
        for summary in summaries:
            summary_start_s = int(date_time_utils.microseconds_to_seconds(summary.start_timestamp_us))
            summary_end_s = int(date_time_utils.microseconds_to_seconds(summary.end_timestamp_us))
            values = numpy.full(len(timestamps), numpy.nan)
            values[timestamps.searchsorted(summary_start_s):timestamps.searchsorted(summary_end_s)] = i
            axes.plot(timestamps, values)

        yticks.append(i)
        ylabels.append(redvox_id.split(":")[0])
        i += 1

    xticks = _subsample(timestamps, 6)
    datetimes = map(datetime.datetime.utcfromtimestamp, xticks)
    datetimes = map(lambda dt: dt.strftime("%d %H:%M"), datetimes)
    axes.set_xticks(xticks)
    axes.set_xticklabels(list(datetimes))
    axes.set_yticks(yticks)
    axes.set_yticklabels(ylabels)

    axes.tick_params(axis='both', which='major', labelsize=8)
    axes.tick_params(axis='both', which='minor', labelsize=7)

    axes.set_title("RedVox Device Summary")
    axes.set_xlabel("Date and Time")
    axes.set_ylabel("Device Activity")

    fig.text(.5, .01, "UTC %s - %s" % (datetime.datetime.utcfromtimestamp(start_s),
                                       datetime.datetime.utcfromtimestamp(end_s)), ha='center')
    plt.show()



"""
This module contains classes and functions for interacting with the RedVox Cloud API authentication endpoints.
"""

from dataclasses import dataclass
from typing import Callable, Optional

from dataclasses_json import dataclass_json
import requests

from redvox.cloud.api import ApiConfig, post_req
from redvox.cloud.routes import RoutesV1


@dataclass_json
@dataclass
class AuthReq:
    """
    An authentication request.
    """
    email: str
    password: str


@dataclass_json
@dataclass
class AuthResp:
    """
    An authentication response.
    """
    status: int
    auth_token: Optional[str]
    claims: Optional['ValidateTokenResp']

    def is_success(self) -> bool:
        return self.status == 200 and self.auth_token is not None and len(self.auth_token) > 0


@dataclass_json
@dataclass
class ValidateTokenReq:
    """
    A token validation request.
    """
    auth_token: str


@dataclass_json
@dataclass
class ValidateTokenResp:
    """
    A verified token response.
    """
    aud: str
    exp: str
    iat: str
    iss: str
    nbf: str
    sub: str
    tier: str


@dataclass_json
@dataclass
class RefreshTokenReq:
    """
    A token validation request.
    """
    auth_token: str


@dataclass_json
@dataclass
class RefreshTokenResp:
    """
    A token validation request.
    """
    auth_token: Optional[str]
    claims: Optional[ValidateTokenResp]


def authenticate_user(api_config: ApiConfig,
                      authentication_request: AuthReq,
                      session: Optional[requests.Session] = None,
                      timeout: Optional[float] = None) -> AuthResp:
    """
    Attempts to authenticate a RedVox user.
    :param api_config: Api configuration.
    :param authentication_request: An instance of an authentication request.
    :param session: An (optional) session for re-using an HTTP client.
    :return: An instance of an authentication response.
    """
    # noinspection Mypy
    handle_resp: Callable[[requests.Response], AuthResp] = lambda resp: AuthResp.from_dict(resp.json())
    res: Optional[AuthResp] = post_req(api_config,
                                       RoutesV1.AUTH_USER,
                                       authentication_request,
                                       handle_resp,
                                       session,
                                       timeout)

    return res if res else AuthResp(401, None, None)


def validate_token(api_config: ApiConfig,
                   validate_token_req: ValidateTokenReq,
                   session: Optional[requests.Session] = None,
                   timeout: Optional[float] = None) -> Optional[ValidateTokenResp]:
    """
    Attempt to validate the provided auth token.
    :param api_config: The Api config.
    :param validate_token_req: A validation token req.
    :param session: An (optional) session for re-using an HTTP client.
    :return: A ValidateTokenResp when the token is valid, None otherwise.
    """
    # noinspection Mypy
    handle_resp: Callable[[requests.Response], ValidateTokenResp] = lambda resp: ValidateTokenResp.from_dict(
        resp.json())
    return post_req(api_config,
                    RoutesV1.VALIDATE_TOKEN,
                    validate_token_req,
                    handle_resp,
                    session,
                    timeout)


def refresh_token(api_config: ApiConfig,
                  refresh_token_req: RefreshTokenReq,
                  session: Optional[requests.Session] = None,
                  timeout: Optional[float] = None) -> Optional[RefreshTokenResp]:
    """
    Attemp to refresh the given authentication token.
    :param api_config: The Api config.
    :param refresh_token_req: The request.
    :param session: An (optional) session for re-using an HTTP client.
    :return: An instance of a RefreshTokenResp.
    """
    # noinspection Mypy
    handle_resp: Callable[[requests.Response], RefreshTokenResp] = lambda resp: RefreshTokenResp.from_dict(resp.json())
    return post_req(api_config,
                    RoutesV1.REFRESH_TOKEN,
                    refresh_token_req,
                    handle_resp,
                    session,
                    timeout)



"""
This module provides functions for gracefully deprecating functionality in this SDK.
"""

import functools
import logging
import traceback
import typing

LOGGER = logging.getLogger("DeprecationWarningLogger")


def deprecated(in_version: str, alt_fn: typing.Callable) -> typing.Callable:
    """
    Decorator used for deprecating functions in the SDK.
    This decorator will display a warning of the deprecated function being called, where it was called from, and then
    call the alternate function.
    :param in_version: Version that the function was deprecated in.
    :param alt_fn: An alternate function to use instead of the deprecated one.
    """
    def decorator_repeat(func):
        """
        Wraps a function.
        :param func: Function to wrap.
        :return: A decorator.
        """
        warn = "%s has been deprecated in version %s. Use %s instead." % (
            func.__name__,
            in_version,
            alt_fn.__name__)

        @functools.wraps(func)
        def wrapper_decorator(*args, **kwargs):
            """
            Displays the deprecation warning and runs the alternate function.
            :param args: Original args.
            :param kwargs: Original kwargs.
            :return: The result of running the alternate function.
            """
            _traceback = traceback.extract_stack(limit=2)[0]
            LOGGER.warning("RedVox SDK Deprecation Warning: %s", warn)
            LOGGER.warning(" -- %s %d %s", _traceback.filename, _traceback.lineno, _traceback.name)
            return alt_fn(*args, **kwargs)

        wrapper_decorator.__doc__ = warn
        return wrapper_decorator

    return decorator_repeat



"""
Support for computing statistics
Requires numpy
"""

from typing import List, Union

import numpy as np


class StatsContainer:
    """
    Helper class to compute statistics for a set of objects
    Stores the mean, std dev, number of data points (count), and best value per set object
    Calculates mean of means, mean of variance, variance of means, total variance,
    and total std dev for the set of objects
    Properties:
        mean_array: the mean of each object in the set
        std_dev_array: the std_dev of each object in the set
        count_array: the number of elements in each object in the set
        best_value: the best value to represent the set
        container_id: a string that identifies the StatsContainer
    """

    def __init__(self, container_id: str) -> None:
        """
        Initialize arrays:
        :param container_id: a string describing the container
        """
        self.mean_array: List[Union[float, int]] = []
        self.std_dev_array: List[Union[float, int]] = []
        self.count_array: List[Union[float, int]] = []
        self.best_value: float = 0.0
        self.container_id: str = container_id

    def mean_of_means(self) -> float:
        """
        Compute the mean of means
        :return: mean of all means
        """
        # convert non-numbers to 0s
        counts: np.ndarray = np.nan_to_num(self.count_array)
        # weight each mean by the number of elements in it
        total_means: np.ndarray = np.prod([np.nan_to_num(self.mean_array), counts], axis=0)
        # if sum(counts) is 0, change sum(counts) to 1 to avoid divide by 0 errors
        return np.sum(total_means) / (1 if np.sum(counts) == 0 else np.sum(counts))

    def mean_of_variance(self) -> float:
        """
        Compute mean of variances
        :return: mean of variances
        """
        # convert non-numbers to 0s
        counts: np.ndarray = np.nan_to_num(self.count_array)
        std_devs: np.ndarray = np.nan_to_num(self.std_dev_array)
        # variance is std dev squared, which is then weighted by the number of elements for that variance
        total_vars: np.ndarray = np.prod([counts, std_devs, std_devs], axis=0)
        # if sum(counts) is 0, change sum(counts) to 1 to avoid divide by 0 errors
        return np.sum(total_vars) / (1 if np.sum(counts) == 0 else np.sum(counts))

    def variance_of_means(self) -> float:
        """
        Compute the variance of means
        :return: variance of means
        """
        counts: np.ndarray = np.nan_to_num(self.count_array)
        # get the difference of individual means and total mean
        mean_vars: np.ndarray = np.subtract(np.nan_to_num(self.mean_array), self.mean_of_means())
        # square the differences then weight them by number of elements
        total: np.ndarray = np.prod([mean_vars, mean_vars, counts], axis=0)
        # if sum(counts) is 0, change sum(counts) to 1 to avoid divide by 0 errors
        return np.sum(total) / (1 if np.sum(counts) == 0 else np.sum(counts))

    def total_variance(self) -> float:
        """
        Compute the variation of all elements
        :return: variation of all elements
        """
        # mean of variances + variance of means = total variance
        return self.mean_of_variance() + self.variance_of_means()

    def total_std_dev(self) -> float:
        """
        Compute the std dev of all elements
        :return: std dev of all elements (sqrt of total variance)
        """
        return np.sqrt(self.total_variance())  # std dev is square root of variance

    def add(self, mean: Union[float, int], std_dev: Union[float, int], count: Union[float, int]) -> None:
        """
        Put an element into the arrays
        :param mean: a mean
        :param std_dev: the std dev for the mean
        :param count: how many values were used to calculate the mean
        """
        self.mean_array.append(mean)
        self.std_dev_array.append(std_dev)
        self.count_array.append(count)



"""
This is an auxiliary file where regularly used constants can be stored and exported.
"""

# noinspection Mypy
import numpy as np


GRAVITY_M_PER_S2: float = 9.80665     # acceleration of gravity in m/s^2
AVG_SEA_LEVEL_PRESSURE_KPA = 101.325  # average sea level pressure around the world in kPa
MOLAR_MASS_AIR_KG_PER_MOL = 0.02896   # molar mass of air in kg per mol
STANDARD_TEMPERATURE_K = 288.15       # standard surface temperature in K
UNIVERSAL_GAS_CONSTANT_KG_M2_PER_K_MOL_S2 = 8.3143  # universal gas constant in (kg * m^2) / (K * mol * s^2)
# constant used in barometric formula
#   source: https://www.math24.net/barometric-formula/
MG_DIV_BY_RT = ((MOLAR_MASS_AIR_KG_PER_MOL * GRAVITY_M_PER_S2) /
                (STANDARD_TEMPERATURE_K * UNIVERSAL_GAS_CONSTANT_KG_M2_PER_K_MOL_S2))

EARTH_RADIUS_M = 6367000.0  # earth's estimated radius in meters

KPA_TO_PA: float = 1000.  # kilopascals in one pascal
EPSILON: float = np.finfo(np.float32).eps
PI: float = np.pi
DEG_TO_RAD = PI / 180.0  # converts degrees to radians
RAD_TO_DEG = 180.0 / PI  # converts radians to degrees

SECONDS_PER_DAY: int = 86_400



"""
This module provides methods and classes for requesting bulk RedVox data.
"""

import logging
from typing import List

import redvox.cloud.api as cloud_api
import redvox.cloud.client as cloud_client
import redvox.cloud.data_api as data_api
import redvox.cloud.data_io as data_io

# pylint: disable=C0103
log = logging.getLogger(__name__)


def make_data_req(out_dir: str,
                  protocol: str,
                  host: str,
                  port: int,
                  email: str,
                  password: str,
                  req_start_s: int,
                  req_end_s: int,
                  redvox_ids: List[str],
                  retries: int,
                  secret_token: str) -> bool:
    """
    Makes a data request to the RedVox data_server.
    :param out_dir: The output directory to store downloaded files.
    :param protocol: One of either https or http.
    :param host: The host of the data server.
    :param port: The port of the data server.
    :param email: The email address of the redvox.io user.
    :param password: The password of the redvox.io user.
    :param req_start_s: The request start time as seconds since the epoch.
    :param req_end_s: The request end time as seconds since the epoch.
    :param redvox_ids: A list of RedVox ids.
    :param retries: The number of retries to perform on failed downloads.
    :param secret_token: A shared secret token required for accessing the data service.
    :return: True if this succeeds, False otherwise.
    """
    api_conf: cloud_api.ApiConfig = cloud_api.ApiConfig(protocol,
                                                        host,
                                                        port)
    client: cloud_client.CloudClient = cloud_client.CloudClient(email,
                                                                password,
                                                                api_conf=api_conf,
                                                                secret_token=secret_token)
    data_resp: data_api.DataRangeResp = client.request_data_range(req_start_s, req_end_s, redvox_ids)
    client.close()

    if len(data_resp.signed_urls) == 0:
        log.error("No signed urls returned")
        return False

    data_io.download_files_parallel(data_resp.signed_urls, out_dir, retries)

    return True



"""
This module contains functions for performing RedVox data conversions and displaying the contents of rdvxz files.
"""

import logging
import os.path
from typing import List, Optional

import redvox.api900.reader as reader
import redvox.api900.reader_utils as reader_utils

# pylint: disable=C0103
log = logging.getLogger(__name__)


def to_json(paths: List[str], out_dir: Optional[str] = None) -> bool:
    """
    Converts .rdvxz files to .json files.
    :param paths: Paths of .rdvxz files to convert.
    :param out_dir: An optional output directory (will use input directory by default)
    :return: True if this succeeds, False otherwise
    """
    for path in paths:
        pb_packet = reader.read_file(path)

        if out_dir is not None:
            file_name: str = os.path.basename(path).replace(".rdvxz", ".json")
            new_path = f"{out_dir}/{file_name}"
        else:
            new_path = path.replace(".rdvxz", ".json")

        with open(new_path, "w") as fout:
            fout.write(reader_utils.to_json(pb_packet))

        log.info("Converted %s to %s", path, new_path)

    return True


def to_rdvxz(paths: List[str], out_dir: Optional[str] = None) -> bool:
    """
    Converts .json files to .rdvxz files.
    :param paths: Paths of .json files to convert.
    :param out_dir: An optional output directory (will use input directory by default)
    :return: True if this succeeds, False otherwise
    """
    for path in paths:
        with open(path, "r") as fin:
            json: str = fin.read()

            if out_dir is not None:
                file_name: str = os.path.basename(path).replace(".json", ".rdvxz")
                new_path = f"{out_dir}/{file_name}"
            else:
                new_path = path.replace(".json", ".rdvxz")

            reader_utils.write_file(new_path, reader_utils.from_json(json))

            log.info("Converted %s to %s", path, new_path)

    return True


def print_stdout(paths: List[str]) -> bool:
    """
    Prints the contends of rdvxz files to sdtout.
    :param paths: Paths to .rdvxz files to print.
    :return: True if this succeeds, False otherwise.
    """
    for path in paths:
        print(reader.read_file(path))

    return True



"""
This module provides a command line interface (CLI) for converting, viewing, and downloading RedVox data files.
"""

import argparse
import logging
import os.path
import sys
from typing import Dict, List, Optional

import redvox.cloud.api as cloud_api
import redvox.cloud.client as cloud_client
import redvox.cloud.data_api as data_api
import redvox.cli.conversions as conversions
import redvox.cli.data_req as data_req

# pylint: disable=C0103
log = logging.getLogger(__name__)


def check_path(path: str, path_is_file: bool = True, file_ext: Optional[str] = None) -> bool:
    """
    Checks that the passed in path exists.
    :param path: The path to check.
    :param path_is_file: The path is a file when True or a directory when False.
    :return: True if the path exists, False otherwise.
    """
    if path_is_file:
        return os.path.isfile(path) and (file_ext is None or os.path.basename(path).endswith(file_ext))
    else:
        return os.path.isdir(path)


def check_files(paths: List[str], file_ext: Optional[str] = None) -> bool:
    """
    Checks this given files to determine if they exist.
    :param paths: The paths to check.
    :param file_ext: An optional file extension to filter against.
    :return: True if all paths exist, False otherwise
    """
    invalid_paths: List[str] = list(filter(lambda path: not check_path(path, file_ext=file_ext), paths))
    if len(invalid_paths) > 0:
        log.error("%d invalid paths found", len(invalid_paths))
        for invalid_path in invalid_paths:
            log.error("Invalid path %s", invalid_path)
        return False
    return True


def check_out_dir(out_dir: Optional[str] = None) -> bool:
    """
    Checks if a given directory exists.
    :param out_dir: The directory to check.
    :return: True if it exists, False otherwise.
    """
    if out_dir is not None and not check_path(out_dir, path_is_file=False):
        log.error("out_dir is invalid: %s", out_dir)
        return False
    return True


def determine_exit(status: bool) -> None:
    """
    Determine the exit status and exit the CLI.
    :param status: True will exit with a status of 0 and False will exit with a status of 1.
    """
    if status:
        log.info("Exiting with status = 0")
        sys.exit(0)

    log.error("Exiting with status = 1")
    sys.exit(1)


def to_json_args(args) -> None:
    """
    Wrapper function that calls the to_json conversion.
    :param args: Args from argparse.
    """
    if not check_files(args.rdvxz_paths, ".rdvxz"):
        determine_exit(False)

    if not check_out_dir(args.out_dir):
        determine_exit(False)

    determine_exit(conversions.to_json(args.rdvxz_paths, args.out_dir))


def to_rdvxz_args(args) -> None:
    """
    Wrapper function that calls the to_rdvxz conversion.
    :param args: Args from argparse.
    """
    if not check_files(args.json_paths, ".json"):
        determine_exit(False)

    if not check_out_dir(args.out_dir):
        determine_exit(False)

    determine_exit(conversions.to_rdvxz(args.json_paths, args.out_dir))


def print_stdout_args(args) -> None:
    """
    Wrapper function that calls the print to stdout.
    :param args: Args from argparse.
    """
    if not check_files(args.rdvxz_paths, ".rdvxz"):
        determine_exit(False)

    determine_exit(conversions.print_stdout(args.rdvxz_paths))


def data_req_args(args) -> None:
    """
    Wrapper function that calls the data_req.
    :param args: Args from argparse.
    """
    if not check_out_dir(args.out_dir):
        determine_exit(False)

    determine_exit(data_req.make_data_req(args.out_dir,
                                          args.protocol,
                                          args.host,
                                          args.port,
                                          args.email,
                                          args.password,
                                          args.req_start_s,
                                          args.req_end_s,
                                          args.redvox_ids,
                                          args.retries,
                                          args.secret_token))


def data_req_report(protocol: str,
                    host: str,
                    port: int,
                    email: str,
                    password: str,
                    report_id: str,
                    out_dir: str,
                    retries: int,
                    secret_token: Optional[str] = None) -> bool:
    """
    Uses the built-in cloud based HTTP API to generate a signed URL for report data and then downloads the report data.
    :param protocol: Either http or https.
    :param host: The data service host.
    :param port: The data service port.
    :param email: The email of the RedVox user.
    :param password: The password of the RedVox user.
    :param report_id: The full RedVox report id.
    :param out_dir: The output directory to play the report distribution.
    :param retries: Number of times to attempt to retry the download on failed attempts.
    :param secret_token: The shared secret if utilized by the API server.
    """
    api_config: cloud_api.ApiConfig = cloud_api.ApiConfig(protocol, host, port)
    client = cloud_client.CloudClient(email,
                                      password,
                                      api_conf=api_config,
                                      secret_token=secret_token)
    resp: Optional[data_api.ReportDataResp] = client.request_report_data(report_id)
    client.close()

    if resp:
        resp.download_fs(out_dir, retries)
        return True

    log.error("Response was None")
    return False


def data_req_report_args(args) -> None:
    """
    Wrapper function that calls the data_req_report.
    :param args: Args from argparse.
    """
    if not check_out_dir(args.out_dir):
        determine_exit(False)

    determine_exit(data_req_report(args.protocol,
                                   args.host,
                                   args.port,
                                   args.email,
                                   args.password,
                                   args.report_id,
                                   args.out_dir,
                                   args.retries,
                                   args.secret_token))


def main():
    """
    Entry point into the CLI.
    """
    parser: argparse.ArgumentParser = argparse.ArgumentParser("redvox-cli",
                                                              description="Command line tools for viewing, converting,"
                                                                          " and downloading RedVox data.")
    parser.add_argument("--verbose",
                        "-v",
                        help="Enable verbose logging",
                        action="count",
                        default=0)

    sub_parser = parser.add_subparsers()
    sub_parser.required = True
    sub_parser.dest = "command"

    # rdvxz -> json
    to_json_parser = sub_parser.add_parser("to_json",
                                           help="Convert rdvxz files to json files")
    to_json_parser.add_argument("rdvxz_paths",
                                help="One or more rdvxz files to convert to json files",
                                nargs="+")
    to_json_parser.add_argument("--out_dir",
                                "-o",
                                help="Optional output directory (will use same directory as source files by default)")
    to_json_parser.set_defaults(func=to_json_args)

    # json -> rdvxz
    to_rdvxz_parser = sub_parser.add_parser("to_rdvxz",
                                            help="Convert json files to rdvxz files")
    to_rdvxz_parser.add_argument("json_paths",
                                 help="One or more json files to convert to rdvxz files",
                                 nargs="+")
    to_rdvxz_parser.add_argument("--out_dir",
                                 "-o",
                                 help="Optional output directory (will use same directory as source files by default)")
    to_rdvxz_parser.set_defaults(func=to_rdvxz_args)

    # print
    print_parser = sub_parser.add_parser("print",
                                         help="Print contents of rdvxz files to stdout")
    print_parser.add_argument("rdvxz_paths",
                              help="One or more rdvxz files to print",
                              nargs="+")
    print_parser.set_defaults(func=print_stdout_args)

    # data_req
    data_req_parser = sub_parser.add_parser("data_req",
                                            help="Request bulk RedVox data from the RedVox servers")
    data_req_parser.add_argument("--out_dir",
                                 "-o",
                                 help="The output directory that RedVox files will be written to (default=.)",
                                 default=".")
    data_req_parser.add_argument("--retries",
                                 "-r",
                                 help="The number of times the client should retry getting a file on failure "
                                      "(default=1)",
                                 default=1,
                                 choices=set(range(0, 6)),
                                 type=int)
    data_req_parser.add_argument("--host",
                                 "-H",
                                 help="Data server host (default=redvox.io)",
                                 default="redvox.io")
    data_req_parser.add_argument("--port",
                                 "-p",
                                 type=int,
                                 help="Data server port (default=8080)",
                                 default=8080)
    data_req_parser.add_argument("--protocol",
                                 help="One of either http or https (default https)",
                                 choices=["https", "http"],
                                 default="https")
    data_req_parser.add_argument("--secret_token",
                                 help="A shared secret token provided by RedVox required for accessing the data "
                                      "request service")
    data_req_parser.add_argument("email",
                                 help="redvox.io account email")
    data_req_parser.add_argument("password",
                                 help="redvox.io account password")
    data_req_parser.add_argument("req_start_s",
                                 type=int,
                                 help="Data request start as number of seconds since the epoch UTC")
    data_req_parser.add_argument("req_end_s",
                                 type=int,
                                 help="Data request end as number of seconds since the epoch UTC")
    data_req_parser.add_argument("redvox_ids",
                                 nargs="+",
                                 help="A list of RedVox ids delimited by a space")
    data_req_parser.set_defaults(func=data_req_args)

    # data req report
    data_req_report_parser = sub_parser.add_parser("data_req_report",
                                                   help="Request bulk RedVox data from the RedVox servers")
    data_req_report_parser.add_argument("--out_dir",
                                        "-o",
                                        help="The output directory that RedVox files will be written to (default=.)",
                                        default=".")
    data_req_report_parser.add_argument("--retries",
                                        "-r",
                                        help="The number of times the client should retry getting a file on failure "
                                             "(default=1)",
                                        default=1,
                                        choices=set(range(0, 6)),
                                        type=int)
    data_req_report_parser.add_argument("--host",
                                        "-H",
                                        help="Data server host (default=redvox.io)",
                                        default="redvox.io")
    data_req_report_parser.add_argument("--port",
                                        "-p",
                                        type=int,
                                        help="Data server port (default=8080)",
                                        default=8080)
    data_req_report_parser.add_argument("--protocol",
                                        help="One of either http or https (default https)",
                                        choices=["https", "http"],
                                        default="https")
    data_req_report_parser.add_argument("--secret_token",
                                        help="A shared secret token provided by RedVox required for accessing the data "
                                             "request service")
    data_req_report_parser.add_argument("email",
                                        help="redvox.io account email")
    data_req_report_parser.add_argument("password",
                                        help="redvox.io account password")
    data_req_report_parser.add_argument("report_id",
                                        type=str,
                                        help="The full report id that data is being requested for")
    data_req_report_parser.set_defaults(func=data_req_report_args)

    # Parse the args
    args = parser.parse_args()

    # Setup logging
    log_levels: Dict[int, str] = {
        0: "WARN",
        1: "INFO",
        2: "DEBUG"
    }
    log_level: str = log_levels[args.verbose] if args.verbose in log_levels else log_levels[0]
    logging.basicConfig(level=log_level,
                        format="[%(levelname)s:%(process)d:%(filename)s:%(module)s:%(funcName)s:%(lineno)d:%(asctime)s]"
                               " %(message)s")

    log.info("Running with args=%s and log_level=%s",
             str(args),
             log_level)

    # Try calling the appropriate handler
    # pylint: disable=W0703
    try:
        args.func(args)
    except Exception as e:
        log.error("Encountered an error: %s", str(e))
        sys.exit(1)


if __name__ == "__main__":
    main()



"""
finds a station's best location and compare it against a surveyed point if one is provided
station location data can be loaded from rdvxz files or input manually
The surveyed point can come from Google earth or any other positioning tool.

PLEASE NOTE:
latitude and longitude measurements are always in degrees (deg)
altitude and accuracy measurements are always in meters (m)
barometer measurements are always in kiloPascals (kPa)
exceptions to this will ALWAYS be noted in comments and variable names

barometric formula source: https://www.math24.net/barometric-formula/
barometric formula P(h) = P0 * e**(h * (-Mg/RT))
where h is a height in meters, P(h) is pressure in kPa at h and P0 is sea-level pressure in kPa
 Mg/RT is a constant based on assumptions of average earth based values.

Haversine equation constants from site: https://movable-type.co.uk/scripts/gis-faq-5.1.html
"""
import pandas as pd
import numpy as np
from typing import List, Dict, Optional, Tuple
from fastkml import kml, styles
from fastkml.geometry import Point
from redvox.api900 import reader
from redvox.common.constants import EPSILON, DEG_TO_RAD, AVG_SEA_LEVEL_PRESSURE_KPA, EARTH_RADIUS_M, \
    STANDARD_TEMPERATURE_K, MOLAR_MASS_AIR_KG_PER_MOL, GRAVITY_M_PER_S2, UNIVERSAL_GAS_CONSTANT_KG_M2_PER_K_MOL_S2

# instruments have only so much accuracy, so if something has a distance less than the following values
#  from a given point, we could feasibly consider it to be close enough to be at the given point.
# default horizontal distance in meters for something to be included with a given point
DEFAULT_INCLUSION_HORIZONTAL_M = 100.0
# default vertical distance in meters for something to be included within a given point
DEFAULT_INCLUSION_VERTICAL_M = 50.0
# default vertical distance in meters computed via barometer measurements to be included within a given point
DEFAULT_INCLUSION_VERTICAL_BAR_M = 10.0

# Survey dictionary minimum keys
# lat: latitude, lon: longitude, alt: altitude, bar: barometer reading
SURVEY_KEYS = ["lat", "lon", "alt", "bar"]
# Optional survey dictionary keys
# sea_bar: pressure reading at sea level
OPTIONAL_SURVEY_KEYS = ["sea_bar"]
# GPS data frame indices
GPS_DATA_INDICES = ["latitude", "longitude", "altitude", "accuracy"]
# closest gps data point to surveyed data frame columns
CLOSEST_TO_SURVEY_COLUMNS = ["closest acc", "closest lat", "closest lon", "closest alt", "closest bar", "distance"]
# mean location data frame columns
MEAN_LOC_COLUMNS = ["mean acc", "mean lat", "mean lon", "mean alt", "mean bar"]
# standard deviation (std) values data frame columns
STD_LOC_COLUMNS = ["std acc", "std lat", "std lon", "std alt", "std bar"]
# station info data frame columns
STATION_INFO_COLUMNS = ["os", "sample rate"]
# master data frame columns
MASTER_COLUMNS = STATION_INFO_COLUMNS + CLOSEST_TO_SURVEY_COLUMNS + MEAN_LOC_COLUMNS + STD_LOC_COLUMNS
# dict of validation methods that can be utilized
VALIDATION_METHODS = {"sol": "close to solution", "mean": "close to mean"}


class DataHolder:
    """
    Stores an array of float data.  The data is privatized for security.
    It also keeps track of the "best value" of the data set.
    Properties:
        id: a string identifier for the data
        _data: private data storage; all values must be floats
        best_value: the value that best represents the data set
    """
    #
    def __init__(self, name: str):
        """
        sets up the DataHolder
        :param name: a string identifier for the data
        """
        self.id = name
        self._data = []
        self.best_value = None

    def add(self, new_data: float):
        """
        adds one element to the data
        :param new_data: float value to add
        """
        self._data.append(new_data)
        self.replace_zeroes_with_epsilon()

    def set_data(self, new_data: List[float]):
        """
        overwrites the stored data with the new_data
        :param new_data: the new list of floats to overwrite the existing data with
        """
        self._data = new_data
        self.replace_zeroes_with_epsilon()

    def replace_zeroes_with_epsilon(self):
        """
        replaces all 0 values in the data with extremely tiny values
        """
        for index in range(len(self._data)):
            if self._data[index] == 0.0:
                self._data[index] = EPSILON

    def get_mean(self) -> float:
        """
        return the mean of the data
        :return: the mean of the data
        """
        return np.mean(self._data)

    def get_std(self) -> float:
        """
        return the standard deviation of the data
        :return: the standard deviation of the data
        """
        return np.std(self._data)

    def get_data(self) -> List[float]:
        """
        return the data
        :return: the data
        """
        return self._data

    def get_len_data(self) -> int:
        """
        return the length of the data array
        :return: the length of the data array
        """
        return len(self._data)


class GPSDataHolder:
    """
    holds gps data (latitude, longitude, altitude, and accuracy) and barometric data
    uses a dataframe to organize the gps data
    Properties:
        gps_df: a dataframe to hold all the gps data
        barometer: a DataHolder for barometer data
        id: string identifier for the data set
        os_type: string identifier for the operating system of the data set
        mic_samp_rate_hz: float sample rate of station microphone in hz
        best_data_index: the index that corresponds to the best representative of the data
    """
    def __init__(self, name: str, opsys: str, data: Optional[List[List[float]]] = None,
                 mic_samp_rate_hz: float = 80., bar: Optional[DataHolder] = None):
        """
        sets up the GPSDataHolder
        :param name: string identifier for the data set
        :param opsys: string identifier for the data set's operating system
        :param data: the data as a list of list of floats, default None
        :param mic_samp_rate_hz: float sample rate of the microphone in hz, default 80 hz
        :param bar: barometer DataHolder, default None
        """
        self.gps_df = pd.DataFrame(data, index=GPS_DATA_INDICES)
        self.barometer = bar
        self.id = name
        self.os_type = opsys
        self.mic_samp_rate_hz = mic_samp_rate_hz
        self.best_data_index = 0

    def clone(self):
        """
        return a copy of the GPSDataHolder
        :return: an exact copy of the GPSDataHolder
        """
        # return a copy of the calling data frame
        new_gps_dh = GPSDataHolder(self.id, self.os_type, None, self.mic_samp_rate_hz, self.barometer)
        new_gps_dh.gps_df = self.gps_df
        new_gps_dh.best_data_index = self.best_data_index
        return new_gps_dh

    def set_data(self, new_data: Optional[List[List[float]]] = None):
        """
        set gps location data.  data is expected to be 4 lists: latitude values, longitude values, altitude values,
          and accuracy values
        :param new_data: list of list of floats that represent the gps data, default None
        """
        self.gps_df = pd.DataFrame(new_data, index=GPS_DATA_INDICES)

    def set_metadata(self, new_id: Optional[str] = None, new_os: Optional[str] = None,
                     new_mic_samp_rate_hz: Optional[float] = None):
        """
        set metadata fields; id, os and mic sample rate
        :param new_id: the new string identifier for the data set, default None
        :param new_os: the new string identifier for the data set's os, default None
        :param new_mic_samp_rate_hz: float of new microphone sample rate in hz, default None
        """
        if new_id is not None:
            self.id = new_id
        if new_os is not None:
            self.os_type = new_os
        if new_mic_samp_rate_hz is not None:
            self.mic_samp_rate_hz = new_mic_samp_rate_hz

    def get_mean_all(self) -> Dict[str, float]:
        """
        return the means of the latitude, longitude, altitude, accuracy, and barometer
        :return: means of all 5 measurements
        """
        bar_mean = self.barometer.get_mean()
        if bar_mean == 0 or bar_mean is None:
            bar_mean = 0.00000000001
        lat_mean = self.gps_df.loc["latitude"].mean()
        lon_mean = self.gps_df.loc["longitude"].mean()
        alt_mean = self.gps_df.loc["altitude"].mean()
        acc_mean = self.gps_df.loc["accuracy"].mean()

        return {"acc": acc_mean, "lat": lat_mean, "lon": lon_mean, "alt": alt_mean, "bar": bar_mean}

    def get_std_all(self) -> Dict[str, float]:
        """
        return the standard deviations of the latitude, longitude, altitude, accuracy, and barometer
        :return: standard deviations of all 5 measurements
        """
        bar_std = self.barometer.get_std()
        if bar_std == 0 or bar_std is None:
            bar_std = 0.00000000001
        lat_std = self.gps_df.loc["latitude"].std()
        if np.isnan(lat_std):
            lat_std = 0
        lon_std = self.gps_df.loc["longitude"].std()
        if np.isnan(lon_std):
            lon_std = 0
        alt_std = self.gps_df.loc["altitude"].std()
        if np.isnan(alt_std):
            alt_std = 0
        acc_std = self.gps_df.loc["accuracy"].std()
        if np.isnan(acc_std):
            acc_std = 0

        return {"acc": acc_std, "lat": lat_std, "lon": lon_std, "alt": alt_std, "bar": bar_std}

    def set_barometer(self, bar_data: List[float]):
        """
        sets the barometer DataHolder.  uses the mean of the data as the best value
        :param bar_data: list of floats to set barometer data as
        """
        self.barometer = DataHolder("barometer")
        self.barometer.set_data(bar_data)
        self.barometer.best_value = np.mean(bar_data)

    def get_size(self) -> (int, int):
        """
        return the amount of gps and barometer data points
        :return: the amount of gps and barometer data points
        """
        return self.gps_df.iloc[0].size, self.barometer.get_len_data()


class LocationAnalyzer:
    """
    stores location information, which can be analyzed later
    contains functions to find mean, standard deviation (std) and validation of data
    use one analyzer per real location point.  one analyzer can accommodate multiple stations per survey point
    the real location dictionary must contain the minimum keys listed in SURVEY_KEYS
    the real location dictionary may contain other keys than ones listed in SURVEY_KEYS
    keys in OPTIONAL_SURVEY_KEYS have special meaning and can only be used as this program intends to use them
    uses dataframes with station id as the index
    Properties:
        all_stations_info_df: dataframe with metadata about all stations
        all_stations_mean_df: dataframe with means from all stations
        all_stations_std_df: dataframe with stds from all stations
        all_stations_closest_df: dataframe with the closest point to the real location and its distance to the
                                    real location for all stations
        invalid_points: a list of gps points that are blacklisted
        _real_location: the surveyed point that the station is located at.  privatized for security
        all_gps_data: a list of all GPSDataHolders that form the data set
        valid_gps_data: a list of all GPSDataHolders that pass validation checks
    """
    def __init__(self, wrapped_packets: List[List[reader.WrappedRedvoxPacket]] = None,
                 real_location: Optional[Dict[str, float]] = None,
                 invalid_points: Optional[List[Dict[str, float]]] = None):
        """
        set up the LocationAnalyzer
        :param wrapped_packets: a list of wrapped redvox packet lists to analyze, default None
        :param real_location: dictionary containing the real location of the station, default None
        :param invalid_points: list of gps points that should not be in the data set, default None
        """
        self.all_stations_closest_df = pd.DataFrame([], columns=CLOSEST_TO_SURVEY_COLUMNS)
        self.all_stations_mean_df = pd.DataFrame([], columns=MEAN_LOC_COLUMNS)
        self.all_stations_std_df = pd.DataFrame([], columns=STD_LOC_COLUMNS)
        self.all_stations_info_df = pd.DataFrame([], columns=STATION_INFO_COLUMNS)
        self.invalid_points = invalid_points
        self.all_gps_data = []
        self.valid_gps_data = []
        self._real_location = real_location
        # if given a path to redvox data, load data from there
        if wrapped_packets is not None:
            for wrapped_device_packets in wrapped_packets:
                self.get_loc_from_packets(wrapped_device_packets)

    def set_real_location(self, survey: Dict[str, float] = None):
        """
        set the real location
        :param survey: dictionary containing the station's location, default None
        """
        self._real_location = survey

    def get_real_location(self) -> Dict[str, float]:
        """
        return the station's real location
        :return: the station's real location
        """
        return self._real_location

    def get_all_dataframes(self) -> pd.DataFrame:
        """
        fuse all dataframes together, joined by station id
        :return: all 4 dataframes fused together
        """
        frames = [self.all_stations_info_df, self.all_stations_closest_df,
                  self.all_stations_mean_df, self.all_stations_std_df]
        return pd.concat(frames, axis=1)

    def get_stats_dataframes(self) -> pd.DataFrame:
        """
        fuse informational and statistic dataframes together, joined by station id
        :return: station info, mean and std dataframes fused together
        """
        frames = [self.all_stations_info_df, self.all_stations_mean_df, self.all_stations_std_df]
        return pd.concat(frames, axis=1)

    def get_loc_from_packets(self, w_p: List[reader.WrappedRedvoxPacket]):
        """
        store the location information and their mean and std using a collection of wrapped redvox packets
        assumes a list of redvox packets shares 1 device id
        :param w_p: a list of wrapped redvox packets to read
        """
        # extract the information from the packets
        sample_rate = w_p[0].microphone_sensor().sample_rate_hz()
        dev_os_type = w_p[0].device_os()
        idd = w_p[0].redvox_id()
        packet_gps_data = load_position_data(w_p)
        # compute mean location
        mean_loc = packet_gps_data.get_mean_all()
        std_loc = packet_gps_data.get_std_all()
        # store the information
        self.all_gps_data.append(packet_gps_data)
        self.all_stations_info_df.loc[idd] = [dev_os_type, sample_rate]
        self.all_stations_std_df.loc[idd] = [std_loc["acc"], std_loc["lat"], std_loc["lon"],
                                             std_loc["alt"], std_loc["bar"]]
        self.all_stations_mean_df.loc[idd] = [mean_loc["acc"], mean_loc["lat"], mean_loc["lon"],
                                              mean_loc["alt"], mean_loc["bar"]]

    def analyze_data(self, write_output: bool = False):
        """
        analyze data, then if a real location exists, compare data to real location
          output is written if enabled
        :param write_output: boolean to write any debugging output, default False
        """
        self.validate_all()
        # if there's no real location, make the mean the real location
        if self._real_location is None:
            means = self.all_stations_mean_df
            self._real_location = {"lat": np.mean(means["mean lat"]), "lon": np.mean(means["mean lon"]),
                                   "alt": np.mean(means["mean alt"]), "bar": np.mean(means["mean bar"])}
        self.compare_with_real_location()
        # print results
        if write_output:
            self.print_to_csv("temp.csv")

    def get_barometric_heights(self, sea_pressure: float = AVG_SEA_LEVEL_PRESSURE_KPA) -> pd.DataFrame:
        """
        for each station, compute the barometric height using the mean
        :param sea_pressure: the local sea pressure in kPa, default AVG_SEA_LEVEL_PRESSURE_KPA
        :return: a dataframe with the barometric heights in meters and station id as the index
        """
        bar_heights = {}
        data_dict = self.all_stations_mean_df["mean bar"].T.to_dict()
        for index in data_dict.keys():
            bar_heights[index] = compute_barometric_height(data_dict[index], sea_pressure)
        barometric_heights = pd.DataFrame(bar_heights, index=["bar height"], columns=self.all_stations_mean_df.index)
        return barometric_heights.T

    def validate_all(self, validation_ranges: Tuple[float, float, float] = (DEFAULT_INCLUSION_HORIZONTAL_M,
                                                                            DEFAULT_INCLUSION_VERTICAL_M,
                                                                            DEFAULT_INCLUSION_VERTICAL_BAR_M), ):
        """
        check that all data in the data set are valid.  Remove outliers and strange values
        :param validation_ranges: tuple of floats that the data values are compared against for validation
        """
        # validation always assumes nothing is valid when it starts, so empty out the valid_gps_data
        self.valid_gps_data = []
        for station in self.all_gps_data:
            # if self._real_location is not None:
            #     validated_gps = self.validator(station, self._real_location)
            # else:
            #     validated_gps = validate_data(station)
            validated_gps = validate(station, validation_ranges, "blacklist", self.invalid_points)
            if validated_gps.get_size()[0] != 0:
                self.valid_gps_data.append(validated_gps)

    def compare_with_real_location(self):
        """
        find the closest valid data point to the real location.  information is stored in the data frames
        """
        # compute closest point to real location
        result = compute_distance_all(self._real_location, self.valid_gps_data)
        self.all_stations_closest_df = result[CLOSEST_TO_SURVEY_COLUMNS]
        self.all_stations_info_df = result[STATION_INFO_COLUMNS]
        self.all_stations_mean_df = result[MEAN_LOC_COLUMNS]
        self.all_stations_std_df = result[STD_LOC_COLUMNS]

    def print_location_df(self, info_type: Optional[str] = None, os_type: Optional[str] = None):
        """
        print a single dataframe or a group of dataframes
        :param info_type: string denoting the type or group of dataframes to output, default None
        :param os_type: string denoting the os of the stations to output, default None
        """
        if info_type == "real":
            print_station_df(self.all_stations_closest_df, os_type)
        elif info_type == "info":
            print_station_df(self.all_stations_info_df, os_type)
        elif info_type == "std":
            print_station_df(self.all_stations_std_df, os_type)
        elif info_type == "mean":
            print_station_df(self.all_stations_mean_df, os_type)
        elif info_type == "all":
            print_station_df(self.get_all_dataframes(), os_type)
        else:
            # fuse statistical dataframes together
            print_station_df(self.get_stats_dataframes(), os_type)

    def print_to_csv(self, path: str, os_type: Optional[str] = None, debug: Optional[bool] = False):
        """
        print dataframes to csv files in path
        :param path: string containing full path and file name
        :param os_type: string denoting the os of the stations to output, default None
        :param debug: if true, output debug statements, default False
        """
        # fuse all dataframes together
        result = self.get_all_dataframes()
        if os_type == "Android":
            get_all_android_station(result).to_csv(path)
        elif os_type == "iOS":
            get_all_ios_station(result).to_csv(path)
        else:
            os_type = "all"
            result.to_csv(path)
        if debug:
            print("Printed {} station data to {}.".format(os_type, path))


def get_all_ios_station(station_df: pd.DataFrame) -> pd.DataFrame:
    """
    return all iOS stations in the dataframe
    :param station_df: the dataframe to search
    :return: a dataframe with all data related to iOS stations in the dataframe
    """
    return station_df.loc[station_df["os"] == "iOS"]


def get_all_android_station(station_df: pd.DataFrame) -> pd.DataFrame:
    """
    return all android stations in the dataframe
    :param station_df: the dataframe to search
    :return: a dataframe with all data related to android stations in the dataframe
    """
    return station_df.loc[station_df["os"] == "Android"]


def print_station_df(station_df: pd.DataFrame, os_type: Optional[str] = None):
    """
    print a dataframe, filtering on the station's os type
    :param station_df: a dataframe to search
    :param os_type: os type to filter on, default None
    """
    if os_type == "Android":
        print(get_all_android_station(station_df))
    elif os_type == "iOS":
        print(get_all_ios_station(station_df))
    else:
        print(station_df)


def load_position_data(w_p: List[reader.WrappedRedvoxPacket]) -> GPSDataHolder:
    """
    load gps data from a list of wrapped redvox packets
    :param w_p: list of wrapped packets to read
    :return: all gps data from the packets in a GPSDataHolder
    """
    gps_data = [[], [], [], []]
    packet = None
    packet_name = None
    bar_data = []
    try:
        for packet in w_p:
            packet_name = packet.default_filename()
            if packet.has_barometer_sensor():
                bar_chan = packet.barometer_sensor()  # load barometer data
                bar_data.extend(bar_chan.payload_values())
            else:
                # add defaults
                bar_data.extend([0.0])
                print("WARNING: {} Barometer empty, using default values!".format(packet_name))
            if packet.has_location_sensor():
                # load each channel's data into the container
                loc_chan = packet.location_sensor()
                gps_data[0].extend(loc_chan.payload_values_latitude())
                gps_data[1].extend(loc_chan.payload_values_longitude())
                gps_data[2].extend(loc_chan.payload_values_altitude())
                gps_data[3].extend(loc_chan.payload_values_accuracy())
            else:
                # add defaults
                gps_data[0].extend([0.0])
                gps_data[1].extend([0.0])
                gps_data[2].extend([0.0])
                gps_data[3].extend([0.0])
                print("WARNING: {} Location empty, using default values!".format(packet_name))
    except Exception as eror:
        if packet is not None:
            error_string = "Something went wrong while reading location data from file: {}.  " \
                           "Original message: {}"
            raise Exception(error_string.format(packet_name, eror))
        else:
            raise Exception("No packet found in file.  Original message: {}".format(eror))

    # load data into data holder
    redvox_id = w_p[0].redvox_id()
    gps_dfh = GPSDataHolder(str(redvox_id), w_p[0].device_os(), gps_data,
                            w_p[0].microphone_sensor().sample_rate_hz())
    gps_dfh.set_barometer(bar_data)

    return gps_dfh


def compute_barometric_height(barometric_pressure: float, sea_pressure: float = AVG_SEA_LEVEL_PRESSURE_KPA,
                              standard_temp: float = STANDARD_TEMPERATURE_K,
                              molar_air_mass: float = MOLAR_MASS_AIR_KG_PER_MOL,
                              gravity: float = GRAVITY_M_PER_S2,
                              gas_constant: float = UNIVERSAL_GAS_CONSTANT_KG_M2_PER_K_MOL_S2) -> float:
    """
    compute height of a single point using a station's barometric and sea-level pressure
    barometric equation from https://www.math24.net/barometric-formula/
    :param barometric_pressure: pressure at a station in kPa
    :param sea_pressure: pressure at sea level in kPa, default AVG_SEA_LEVEL_PRESSURE_KPA
    :param standard_temp: surface temperature in K, default STANDARD_TEMPERATURE_K
    :param molar_air_mass: molar mass of air in kg/mol, default MOLAR_MASS_AIR_KG_PER_MOL
    :param gravity: the acceleration of gravity in m/s2, default GRAVITY_M_PER_S2
    :param gas_constant: the universal gas constant in (kg * m2)/(K * mol * s2),
                            default UNIVERSAL_GAS_CONSTANT_KG_M2_PER_K_MOL_S2
    :return: height of station in meters
    """
    # formula and derivations:
    # P(h) = P0 * e**(-Mgh/RT) where:
    # P0 = AVG_SEA_LEVEL_PRESSURE_KPA = 101.325
    # g = GRAVITY_M_PER_S2 = 9.807
    # M = MOLAR_MASS_AIR_KG_PER_MOL = 0.02896
    # T = STANDARD_TEMPERATURE_K = 288.15
    # R = UNIVERSAL_GAS_CONSTANT_KG_M2_PER_K_MOL_S2 = 8.3143
    # therefore h = ln(P0/P(h)) / (Mg/RT)
    # due to log function, we can't let sea_pressure or barometric_pressure be 0
    if sea_pressure == 0.0:
        sea_pressure = EPSILON
    if barometric_pressure == 0.0:
        barometric_pressure = EPSILON
    barometric_height = np.log(sea_pressure / barometric_pressure) / (
            (molar_air_mass * gravity) / (standard_temp * gas_constant))
    return barometric_height


def compute_barometric_height_array(barometric_pressure: np.array, sea_pressure: float = AVG_SEA_LEVEL_PRESSURE_KPA,
                                    standard_temp: float = STANDARD_TEMPERATURE_K,
                                    molar_air_mass: float = MOLAR_MASS_AIR_KG_PER_MOL,
                                    gravity: float = GRAVITY_M_PER_S2,
                                    gas_constant: float = UNIVERSAL_GAS_CONSTANT_KG_M2_PER_K_MOL_S2) -> np.array:
    """
    compute height of many points using each station's barometric and sea-level pressure
    :param barometric_pressure: array of pressures at stations in kPa
    :param sea_pressure: pressure at sea level in kPa, default AVG_SEA_LEVEL_PRESSURE_KPA
    :param standard_temp: surface temperature in K, default STANDARD_TEMPERATURE_K
    :param molar_air_mass: molar mass of air in kg/mol, default MOLAR_MASS_AIR_KG_PER_MOL
    :param gravity: the acceleration of gravity in m/s2, default GRAVITY_M_PER_S2
    :param gas_constant: the universal gas constant in (kg * m2)/(K * mol * s2),
                            default UNIVERSAL_GAS_CONSTANT_KG_M2_PER_K_MOL_S2
    :return: the height of each station in meters
    """
    # due to log function, we can't let sea_pressure or barometric_pressure be 0
    if sea_pressure == 0.0:
        sea_pressure = EPSILON
    for index in range(len(barometric_pressure)):
        if barometric_pressure[index] == 0.0:
            barometric_pressure[index] = EPSILON
    barometric_height = np.log(sea_pressure / barometric_pressure) / (
            (molar_air_mass * gravity) / (standard_temp * gas_constant))
    return barometric_height


def get_component_dist_to_point(point: Dict[str, float], gps_data: pd.Series, bar_mean: float) -> \
        (float, float, float):
    """
    compute distance from the gps data point to the chosen point using haversine formula
    :param point: dict with location to compute distance to
    :param gps_data: series with gps data of one point
    :param bar_mean: the mean barometer reading
    :return: the distance in meters of the horizontal and vertical gps components and barometer readings
    """
    # horizontal distance, use haversine formula
    dlon = gps_data["longitude"] - point["lon"]
    dlat = gps_data["latitude"] - point["lat"]
    haver = np.sin(dlat * DEG_TO_RAD / 2.0) ** 2.0 + (np.cos(point["lat"] * DEG_TO_RAD) *
                                                      np.cos(gps_data["latitude"] * DEG_TO_RAD) *
                                                      np.sin(dlon * DEG_TO_RAD / 2.0) ** 2.0)
    c = 2.0 * np.arcsin(np.min([1.0, np.sqrt(haver)]))
    h_dist = EARTH_RADIUS_M * c
    # vertical distance
    v_dist = np.abs(gps_data["altitude"] - point["alt"])
    # vertical distance using barometer
    v_bar_dist = np.abs(compute_barometric_height(bar_mean) - point["alt"])
    return h_dist, v_dist, v_bar_dist


def get_gps_dist_to_location(point: Dict[str, float], gps_dataholder: GPSDataHolder,
                             bar_alt: Optional[float] = None) -> np.array:
    """
    compute distance from multiple gps points to the chosen point using haversine formula
    :param point: dict with location to compute distance to
    :param gps_dataholder: all the gps data points to compute distance from
    :param bar_alt: height as measured by a barometer, default None
    :return: array of all distances in meters from gps point to chosen point
    """
    # compute distance from the gps data points to the location
    # if given a barometer altitude value, use that instead of the gps altitude
    if bar_alt is not None:
        station_alt = bar_alt
    else:
        station_alt = gps_dataholder.gps_df.loc["altitude"].to_numpy()
    # user haversine formula
    dlon = gps_dataholder.gps_df.loc["longitude"].to_numpy() - point["lon"]
    dlat = gps_dataholder.gps_df.loc["latitude"].to_numpy() - point["lat"]
    haver = np.sin(dlat * DEG_TO_RAD / 2.0) ** 2.0 + (np.cos(point["lat"] * DEG_TO_RAD) *
                                                      np.cos(gps_dataholder.gps_df.loc["latitude"].to_numpy()
                                                             * DEG_TO_RAD)
                                                      * np.sin(dlon * DEG_TO_RAD / 2.0) ** 2.0)
    c = 2 * np.arcsin(np.minimum([1.0], np.sqrt(haver)))
    h_dist = EARTH_RADIUS_M * c
    dist_array = (h_dist ** 2 + (point["alt"] - station_alt) ** 2)
    return np.sqrt(dist_array)


def validate_blacklist(gps_data: pd.Series, point: Dict[str, float], bar_mean: float,
                       inclusion_ranges: Tuple[float, float, float] = (DEFAULT_INCLUSION_HORIZONTAL_M,
                                                                       DEFAULT_INCLUSION_VERTICAL_M,
                                                                       DEFAULT_INCLUSION_VERTICAL_BAR_M)) -> bool:
    """
    return True if the gps point is NOT in the blacklist
    :param gps_data: data to compare
    :param point: the point that is blacklisted
    :param bar_mean: the mean of the barometer measurements
    :param inclusion_ranges: distance from blacklisted point to be considered close enough
    :return: True if point is not in blacklisted point's vicinity
    """
    # calculate distance from gps data to invalid point
    h_dist, v_dist, v_bar_dist = get_component_dist_to_point(point, gps_data, bar_mean)
    # if outside horizontal and vertical distance, we're far enough away from the invalid point
    return h_dist > inclusion_ranges[0] and (v_dist > inclusion_ranges[1] or v_bar_dist > inclusion_ranges[2])


def validate_near_point(gps_data: pd.Series, point: Dict[str, float], bar_mean: float,
                        inclusion_ranges: Tuple[float, float, float] = (DEFAULT_INCLUSION_HORIZONTAL_M,
                                                                        DEFAULT_INCLUSION_VERTICAL_M,
                                                                        DEFAULT_INCLUSION_VERTICAL_BAR_M)) -> bool:
    """
    return True if the gps point IS close to the chosen point
    :param gps_data: data to compare
    :param point: the chosen point to compare against
    :param bar_mean: the mean of the barometer measurements
    :param inclusion_ranges: distance from chosen point to be considered close enough
    :return: True if point is within the chosen point's vicinity
    """
    # calculate distance from gps data to point
    h_dist, v_dist, v_bar_dist = get_component_dist_to_point(point, gps_data, bar_mean)
    # if within horizontal distance and vertical distance, we're close enough to the point
    return h_dist <= inclusion_ranges[0] and (v_dist <= inclusion_ranges[1] or v_bar_dist <= inclusion_ranges[2])


def point_on_line_side(line_points: Tuple[Dict[str, float], Dict[str, float]], point: Dict[str, float]) -> float:
    """
    check which side of a line the point is on
    algorithm from: http://geomalgorithms.com/a03-_inclusion.html
    :param line_points: two coordinates that define a line
    :param point: point to test
    :return: < 0 for right side, == 0 for on line, > 0 for left side
    """
    return ((line_points[1]["lon"] - line_points[0]["lon"]) * (point["lat"] - line_points[0]["lat"]) -
            (point["lon"] - line_points[0]["lon"]) * (line_points[1]["lat"] - line_points[0]["lat"]))


def validate_point_in_polygon(point: Dict[str, float], edges: List[Dict[str, float]]) -> bool:
    """
    Use winding number algorithm to determine if a point is in a polygon (or on the edge)
    if winding number is 0, point is outside polygon
    algorithm from: http://geomalgorithms.com/a03-_inclusion.html
    :param point: coordinates of the point to compare
    :param edges: list of coordinates of the edges of the polygon, with the last edge equal to the first
    :return: True if point is in the polygon
    """
    wn = 0  # winding number
    for index in range(len(edges) - 1):
        if edges[index]["lat"] <= point["lat"]:
            if edges[index + 1]["lat"] > point["lat"]:
                if point_on_line_side((edges[index], edges[index + 1]), point) >= 0:
                    wn += 1
        elif edges[index + 1]["lat"] <= point["lat"]:
            if point_on_line_side((edges[index], edges[index + 1]), point) <= 0:
                wn -= 1
    return wn != 0


def validate(data_to_test: GPSDataHolder,
             inclusion_ranges: Tuple[float, float, float] = (DEFAULT_INCLUSION_HORIZONTAL_M,
                                                             DEFAULT_INCLUSION_VERTICAL_M,
                                                             DEFAULT_INCLUSION_VERTICAL_BAR_M),
             validation_type: str = None, validation_points: List[Dict[str, float]] = None,
             debug: bool = False) -> GPSDataHolder:
    """
    validation master function.  Can perform any kind of validation requested
    :param data_to_test: gps data to validate
    :param inclusion_ranges: ranges to include a data point with a validation point
    :param validation_type: the kind of validation to perform, default None
    :param validation_points: the points to validate against, default None
    :param debug: if True, output debugging information, default False
    :return: all valid gps data
    """
    # perform validation.  returns all valid data
    # check if we even have points to compare against
    if len(validation_points) < 1:
        return data_to_test  # no points to check, everything is good
    need_to_test_gps = data_to_test.clone()
    while True:  # keep going until the data doesn't change
        # remove any points in the data that are not close to the points
        validated_gps_data = [[], [], [], []]
        for gps_point in data_to_test.gps_df.columns:
            # extract data to test
            gps_data = data_to_test.gps_df[gps_point]
            point_valid = True  # assume point is valid
            for point in validation_points:
                if validation_type == "solution" or validation_type == "mean":
                    point_valid = validate_near_point(gps_data, point,
                                                      data_to_test.barometer.get_mean(), inclusion_ranges)
                    if not point_valid:  # if point_valid ever becomes false, we can stop processing
                        break
                else:
                    point_valid = validate_blacklist(gps_data, point,
                                                     data_to_test.barometer.get_mean(), inclusion_ranges)
                    if not point_valid:  # if point_valid ever becomes false, we can stop processing
                        break
            if point_valid:  # add only valid points
                validated_gps_data[0].append(gps_data["latitude"])
                validated_gps_data[1].append(gps_data["longitude"])
                validated_gps_data[2].append(gps_data["altitude"])
                validated_gps_data[3].append(gps_data["accuracy"])
        # create the object to return.
        validated_gps = GPSDataHolder(data_to_test.id, data_to_test.os_type, validated_gps_data,
                                      data_to_test.mic_samp_rate_hz, data_to_test.barometer)
        if validated_gps.get_size() == need_to_test_gps.get_size():
            # print message if user allows it
            if debug:
                print("{} data validated".format(validated_gps.id))
            # if data does not change, we are done validating
            return validated_gps
        else:
            # use the new data to update the old data
            need_to_test_gps = validated_gps.clone()


def compute_distance_all(point: Dict[str, float], all_gps_data: List[GPSDataHolder]) -> pd.DataFrame:
    """
    compute distance from all gps data points to the chosen point
    :param point: the point to compute distance to
    :param all_gps_data: the gps data points to compute distance from
    :return: dataframe containing all information about the gps points' distance to the chosen point
    """
    # compare distances from multiple gps points to the location.  return the closest point
    closeness = {}
    for gps_dh in all_gps_data:
        closeness.update(compute_distance(point, gps_dh))
    return pd.DataFrame(closeness, index=MASTER_COLUMNS).T


def compute_distance(point: dict, gps_data: GPSDataHolder) -> dict:
    """
    compute the distance from the gps points to the chosen point
    :param point: the chosen point to compute distance to
    :param gps_data: the data points to compute distance from
    :return: dictionary containing all information about the gps points' distance to the chosen point
    """
    # for a location, compute distance to closest data point
    idd = gps_data.id
    stations_data = {idd: None}
    gps_loc = gps_data.get_mean_all()

    # find closest barometer altitude to location
    # bar_alt_tmp = (((SEA_PRESSURE / np.array(gps_data.barometer.data)) ** 0.190263096) - 1) * (SOL_TEMP / 0.0065)
    if "sea bar" in point.keys() and point["sea_bar"] is not None:
        bar_alt_tmp = compute_barometric_height_array(np.array(gps_data.barometer.get_data()), point["sea_bar"])
    else:
        bar_alt_tmp = compute_barometric_height_array(np.array(gps_data.barometer.get_data()))
    # simplified barometric equation:
    # P(h) = 101.325 * e ** (-0.00012h) -> P(h) / 101.325 = 1 / (e ** 0.00012h)
    # e ** 0.00012h = 101.325 / P(h) -> 0.00012h = ln(101.325) - ln(P))
    # SEA_PRESSURE = 101.325
    # h = ln(SEA_PRESSURE/P(h)) / 0.00012
    min_index = np.argmin(np.abs(bar_alt_tmp - point["alt"]))
    gps_data.barometer.best_value = gps_data.barometer.get_data()[min_index]
    bar_alt = bar_alt_tmp[min_index]
    # for all gps coords, find closest to solution
    dist_array = get_gps_dist_to_location(point, gps_data)
    min_index = np.argmin(dist_array)
    # compute distance using best barometer measurement
    dist_array_bar = get_gps_dist_to_location(point, gps_data, bar_alt)
    min_bar_index = np.argmin(dist_array_bar)
    # compare minimum of pure gps and gps with barometer
    if dist_array_bar[min_bar_index] < dist_array_bar[min_index]:
        min_index = min_bar_index
        dist_array = dist_array_bar

    # finding the std of the distances is basically finding the std of accuracy
    acc_std = np.std(dist_array)
    lat_std = np.std(np.abs(point["lat"] - gps_data.gps_df.loc["latitude"].to_numpy()))
    lon_std = np.std(np.abs(point["lon"] - gps_data.gps_df.loc["longitude"].to_numpy()))
    alt_std = np.std(np.abs(point["alt"] - gps_data.gps_df.loc["altitude"].to_numpy()))
    bar_std = gps_data.barometer.get_std()

    # put data into dictionary to store in data frames later
    stations_data[idd] = [gps_data.os_type, gps_data.mic_samp_rate_hz, gps_data.gps_df.loc["accuracy", min_index],
                          gps_data.gps_df.loc["latitude", min_index], gps_data.gps_df.loc["longitude", min_index],
                          gps_data.gps_df.loc["altitude", min_index], gps_data.barometer.best_value,
                          dist_array[min_index], gps_loc["acc"], gps_loc["lat"], gps_loc["lon"],
                          gps_loc["alt"], gps_loc["bar"], acc_std, lat_std, lon_std, alt_std, bar_std]

    return stations_data


def load_kml(kml_file: str) -> Dict[str, Dict[str, float]]:
    """
    load location from a kml file
    :param kml_file: full path of the file to load data from
    :return: dictionary of locations with identifiers
    """
    with open(kml_file, 'r', encoding="utf-8") as my_file:
        kml_doc = my_file.read()
    kml_data = kml.KML()
    kml_data.from_string(bytes(kml_doc, encoding="utf8"))
    locations = list(list(kml_data.features())[0].features())
    set_locations = {}
    for place in locations:
        set_locations[place.name] = {"lon": place.geometry.x, "lat": place.geometry.y, "alt": place.geometry.z}
    return set_locations


def write_kml(kml_file: str, master_dict: Dict[str, Dict[str, float]]):
    """
    put information from master_dict into a kml file
    :param kml_file: full path of kml file to write data to
    :param master_dict: the dictionary of information to write
    """
    ns = '{http://www.opengis.net/kml/2.2}'
    # declare kml structure and the document
    kmlz = kml.KML(ns=ns)
    doc = kml.Document(ns, '1')
    # declare, then add styles to doc
    doc_style = styles.Style(id='2')
    pnt_style = styles.IconStyle(id='3', color='ff0000ff')
    pnt_style.icon_href = "http://maps.google.com/mapfiles/kml/shapes/placemark_circle.png"
    doc_style.append_style(pnt_style)
    doc.append_style(doc_style)
    # id is assigned dynamically as new elements are created
    new_id = 4
    for key in master_dict.keys():
        # how do we know if bar is better than alt?
        # set point description to os and sample rate
        description = "{} {}hz".format(master_dict[key]["os"], str(master_dict[key]["sample rate"]))
        # declare the placemark, then give it some coordinates
        pnt = kml.Placemark(ns, id=str(new_id), name=key, description=description, styleUrl='#2')
        new_id += 1
        pnt.geometry = Point(master_dict[key]["mean lon"], master_dict[key]["mean lat"], master_dict[key]["mean alt"])
        # add placemark to doc
        doc.append(pnt)
    # add the doc to the kml file
    kmlz.append(doc)
    # write the kml file, with nice formatting
    with open(kml_file, 'w', encoding="utf-8") as my_file:
        my_file.write('<?xml version="1.0" encoding="UTF-8"?>\n')
        my_file.write(kmlz.to_string(prettyprint=True))
